@@ -16,6 +16,7 @@
  */
 
 struct progress_indicator {
+	const char		*msg;
 	struct bbpos		pos;
 	unsigned long		next_print;
 	u64			nodes_seen;
@@ -24,24 +25,15 @@ struct progress_indicator {
 	bool			silent;
 };
 
-void bch2_progress_init_inner(struct progress_indicator *s,
+void bch2_progress_init(struct progress_indicator *s,
+			      const char *msg,
 			      struct bch_fs *c,
 			      u64 leaf_btree_id_mask,
 			      u64 inner_btree_id_mask);
 
-static inline void bch2_progress_init(struct progress_indicator *s,
-				      struct bch_fs *c, u64 btree_id_mask)
-{
-	bch2_progress_init_inner(s, c, btree_id_mask, 0);
-}
-
 int bch2_progress_update_iter(struct btree_trans *,
 			      struct progress_indicator *,
-			      struct btree_iter *,
-			      const char *);
-
-#define progress_update_iter(trans, p, iter)			\
-	bch2_progress_update_iter(trans, p, iter, __func__)
+			      struct btree_iter *);
 
 void bch2_progress_to_text(struct printbuf *, struct progress_indicator *);
 
